
info :-
        nl,
        write('Usage:'), nl,
        write('?- tell('), list_put("'"),
        write('miik.pat'), list_put("'"),
        write('), go, told.'), nl, nl,
        write('(This may take a while)'), nl.

heading :-
        write('#'), nl,
        write('# FGREP sentence data'), nl,
        write('# Source:'), nl,
        write('#   Risto Miikkulainen and Michael G. Dyer'), nl,
        write('#   Natural Language Processing With Modular'), nl,
        write('#     PDP Networks and Distributed Lexicon'), nl,
        write('#   Cognitive Science 15, p. 343-399 (1991)'), nl,
        write('#   Appendix A: Sentence Data, p. 395-396'), nl,
        write('#'), nl,
        write('# INPUT SENTENCE / AGENT ACT PATIENT INSTRUMENT MODIFIER'), nl.

sio([Human, ate], [Human, ate]) :-
        wrd(Human, human).
sio([Human, ate, Food], [Human, ate, Food]) :-
 	wrd(Human, human),
 	wrd(Food, food).
sio([Human, ate, Food1, Food2], [Human, ate, Food1, '!NILL', Food2]) :-
	wrd(Human, human),
	wrd(Food1, food),
        wrd(Food2, food).
%        Food1 \= Food2.
sio([Human, ate, Food, Utensil], [Human, ate, Food, Utensil]) :-
	wrd(Human, human),
	wrd(Food, food),
	wrd(Utensil, utensil).
sio([Animal, ate], [Animal, ate]) :-
        wrd(Animal, animal).
sio([Predator, ate, Prey], [Predator, ate, Prey]) :-
	wrd(Predator, predator),
	wrd(Prey, prey).
sio([Human, broke, Fragileobj], [Human, broke, Fragileobj]) :-
	wrd(Human, human),
	wrd(Fragileobj, fragileobj).
sio([Human, broke, Fragileobj, Breaker], [Human, broke, Fragileobj, Breaker]) :-
	wrd(Human, human),
	wrd(Fragileobj, fragileobj),
	wrd(Breaker, breaker).
sio([Breaker, broke, Fragileobj], ['!NILL', broke, Fragileobj, Breaker]) :-
	wrd(Breaker, breaker),
	wrd(Fragileobj, fragileobj).
sio([Animal, broke, Fragileobj], [Animal, broke, Fragileobj]) :-
	wrd(Animal, animal),
	wrd(Fragileobj, fragileobj).
sio([Fragileobj, broke], ['!NILL', broke, Fragileobj]) :-
	wrd(Fragileobj, fragileobj).
sio([Human, hit, Thing], [Human, hit, Thing]) :-
	wrd(Human, human),
	wrd(Thing, thing).
sio([Human1, hit, Human2, Possession], [Human1, hit, Human2, '!NILL', Possession]) :-
	wrd(Human1, human),
	wrd(Human2, human),
	wrd(Possession, possession).
sio([Human, hit, Thing, Hitter], [Human, hit, Thing, Hitter]) :-
	wrd(Human, human),
	wrd(Thing, thing),
	wrd(Hitter, hitter).
sio([Hitter, hit, Thing], ['!NILL', hit, Thing, Hitter]) :-
	wrd(Hitter, hitter),
	wrd(Thing, thing).
sio([Human, moved], [Human, moved, Human]) :-
	wrd(Human, human).
sio([Human, moved, Object], [Human, moved, Object]) :-
	wrd(Human, human),
	wrd(Object, object).
sio([Animal, moved], [Animal, moved, Animal]) :-
	wrd(Animal, animal).
sio([Object, moved], ['!NILL', moved, Object]) :-
	wrd(Object, object).

cat(human, [man, woman, boy, girl]).
cat(animal, [bat, chicken, dog, sheep, wolf, lion]).
cat(predator, [wolf, lion]).
cat(prey, [chicken, sheep]).
cat(food, [chicken, cheese, pasta, carrot]).
cat(utensil, [fork, spoon]).
cat(fragileobj, [plate, window, vase]).
cat(hitter, [bat, ball, hatchet, hammer, vase, paperweight, rock]).
cat(breaker, [bat, ball, hatchet, hammer, paperweight, rock]).
cat(possession, [bat, ball, hatchet, hammer, vase, dog, doll]).
cat(object, [bat, ball, hatchet, hammer, paperweight, rock, vase, plate,
        window, fork, spoon, pasta, cheese, chicken, carrot,
        desk, doll, curtain]).
cat(thing, Things) :-
        cat(human, Humans),
        cat(animal, Animals),
        cat(object, Objects),
        conc(Animals, Objects, AnimalsObjects),
        conc(Humans, AnimalsObjects, Things).

wrd(Wrd, Cat) :-
        cat(Cat, Wrds),
        member(Wrd, Wrds).

go :-
        retract(pat(_,_)),
        fail.
go :-
        heading,
        sio(In, Out),
        \+ pat(In, Out),
        assert(pat(In, Out)),
        list_write(In),
        write('/ '),
        list_write(Out),
        nl,
        fail.
go.

list_put([]).
list_put([H | T]) :-
        put(H),
        list_put(T).

list_write([]).
list_write([H | T]) :-
        write(H), tab(1),
        list_write(T).

member(H, [H | _]).
member(X, [_ | T]) :-
        member(X, T).

conc([], L, L).
conc([H | T1], L, [H | T2]) :-
        conc(T1, L, T2).
