
/*  file: miik.c

/*  Forming Global Representations with Extended backPropagation
 *  (FGREP)
 */

#ifdef __MSDOS__
#ifndef __COMPACT__
#error Memory model COMPACT required
#endif /* __COMPACT__ */
#include <alloc.h>
#include <dir.h>
#include <io.h>
#endif /* __MSDOS */
#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef __MSDOS__
#define PATH_SEP '\\'
#else
#define PATH_SEP '/'
#define stricmp(A,B) (strcasecmp(A,B))
#endif

#define BUFSIZE 5000

int
    n_in,               /*  aantal input units                  */
    n_hid,              /*  aantal hidden units                 */
    n_out,              /*  aantal output units                 */
    n_pat,              /*  aantal input/output patronen        */
    *p_vec = NULL,      /*  indirecte pointers naar patronen    */
    all_updated,
    dont_save = 0,
    show_epoch = 1,     /*  aantal epochs tussen uitvoer naar scherm  */
    vec_len = 6,        /*  aantal units per naam               */
    in_vec_len,         /*  aantal namen per input              */
    out_vec_len,        /*  aantal namen per output             */
    **in_vecs = NULL,   /*  input vector als volgnummers van namen  */
    **out_vecs = NULL,  /*  output vector als volgnummers van namen */
    n_names,            /*  aantal vectornamen                  */
    max_names,          /*  maximale aantal vectornamen in tabel vec_names  */
    nill,               /*  index van "!NILL" in vec_names      */
    any,                /*  index van "!ANY" in vec_names       */
    max_epoch = 10000,
    epoch,              /*  aantal trainingen                   */
    verbose = 1;
float
    **p_err = NULL,     /*  fout per patroon                                */
    **p_in = NULL,      /*  inputvectoren                                   */
    **p_out = NULL,     /*  gewenste outputvectoren                         */
    *in = NULL,         /*  invoercellen                                    */
    *hid = NULL,        /*  verborgen cellen                                */
    *out = NULL,        /*  uitvoervector                                   */
    *target = NULL,     /*  gewenste uitvoer                                */
    **w1 = NULL,        /*  gewichten van invoer naar hidden cellen         */
    **w2 = NULL,        /*  gewichten van hidden naar uitvoercellen         */
    *d0 = NULL,         /*  aanpassing inputvectoren                        */
    *d1 = NULL,         /*  aanpassing w1                                   */
    *d2 = NULL,         /*  aanpassing w2                                   */
    eta = .5,           /*  leerconstante                                   */
    randomweight = 0.3, /*  uiterste beginwaarde (-/+) gewichten            */
    randomvector = 0.3, /*  uiterste beginwaarde (-/+) gewichten: sigmoid   */
    **vec_val = NULL,   /*  de waarden van tekstvectoren                    */
    quit_val = 0.05;    /*  einde indien grootste fout kleiner dan dit      */

char
    *programname,
    *no_mem_buffer,
    Out_of_memory [] = "Out of memory",
    *patternfile = NULL,
    *oldweightsfile = NULL,
    *namesfile = NULL,
    buffer [BUFSIZE + 1],
    **vec_names = NULL;         /*  woorden als input/output    */

void
    get_programname (char const *argv0),
    no_mem (void),
    errit (char const *format, ...),
    process_args (int argc, char *argv []),
    read_vector_file (void),
    read_oldweights (void),
    read_namesfile (void),
    *s_malloc (size_t size),
    *s_realloc (void *block, size_t size),
    save_results (int argc, char *argv []),
    random_weights (void),
    feed_forward (int vectornumber),
    rnd_vec (void);
int
    back_propagate (int vectornumber),
    locate (char const *name);
float
    update_all_error (void),
    display_error (void);

char
    *get_arg (int argc, char *argv [], int *index),
    *s_strdup (char const *s),
    *getline (FILE *fp, char const *required);

FILE
    *open_read (char const *filename);

int main (int argc, char *argv [])
{
    int
        vec,
        bp,
        view_lines;

    no_mem_buffer = (char *) malloc (1024);

    srand ((unsigned int) time (NULL));
    process_args (argc, argv);

    vec = n_pat - 1,
    bp = (update_all_error () > quit_val);

    epoch = -1;
    view_lines = 0;
    for (;;) {
        if (++vec == n_pat) {
            epoch++;
            if (verbose && ((view_lines % 24) == 0)) {
                printf (
                    " Epoch   Totaal  Hoogste >%4.2f"
                    "  Dev.W1  Dev.W2  Gem.W1  Gem.W2"
                    " Dev.Vec Gem.Vec\n",
                    quit_val * 3
                );
                view_lines++;
            }
            if (
                (! bp)
             || (epoch == max_epoch)
            ) {
                if (verbose)
                    printf ("\a");
                break;
            }
            if (verbose && ! (epoch % show_epoch)) {
                display_error ();
                view_lines++;
            }
            vec = bp = 0;
            rnd_vec ();
        }

        if (verbose == 2) {
            printf ("%6i%4i\r", epoch, vec);
            fflush (stdout);
        }

        feed_forward (p_vec [vec]);
        if (back_propagate (p_vec [vec]))
            bp = 1;
    }
    update_all_error ();
    if (verbose)
        display_error ();

    save_results (argc, argv);

    return 0;
}

void feed_forward (int vectornumber)
{
    int
        i, j;
    float
        som;

    for (i = 0; i < in_vec_len; i++)
        memcpy (
            in + 1 + i * vec_len,
            vec_val [in_vecs [vectornumber][i]],
            vec_len * sizeof (vec_val [0][0])
        );
    for (i = 0; i < out_vec_len; i++)
        memcpy (
            target + 1 + i * vec_len,
            vec_val [out_vecs [vectornumber][i]],
            vec_len * sizeof (vec_val [0][0])
        );


    for (j = 1; j <= n_hid; j++) {
        som = 0.0;
        for (i = 0; i <= n_in; i++)
            som -= w1 [i][j] * in [i];
        hid [j] = 1.0 / (1.0 + exp (som));
    }
    for (j = 1; j <= n_out; j++) {
        som = 0.0;
        for (i = 0; i <= n_hid; i++)
            som -= w2 [i][j] * hid [i];
        out [j] = 1.0 / (1.0 + exp (som));
    }
}

int back_propagate (int vectornumber)
{
    int
        i, j,
        skip = 1;
    float
        f,
        err,
        som;

    all_updated = 0;

    for (i = 1; i <= n_out; i++) {
        if (
            (
                p_err [vectornumber][i - 1] = fabs (err = target [i] - out [i])
            ) > quit_val
        )
            skip = 0;
        d2 [i] = out [i] * (1 - out [i]) * err;
    }
    if (skip)
        return 0;

    for (i = 1; i <= n_hid; i++) {
        som = 0;
        for (j = 1; j <= n_out; j++)
            som += d2 [j] * w2 [i][j];
        d1 [i] = hid [i] * (1 - hid [i]) * som;
    }

    for (i = 1; i <= n_in; i++) {
        som = 0;
        for (j = 1; j <= n_hid; j++)
            som += d1 [j] * w1 [i][j];
        d0 [i] = in [i] * (1 - in [i]) * som;
    }

    for (i = 0; i <= n_hid; i++)
        for (j = 1; j <= n_out; j++)
            w2 [i][j] += eta * d2 [j] * hid [i];

    for (i = 0; i <= n_in; i++)
        for (j = 1; j <= n_hid; j++)
            w1 [i][j] += eta * d1 [j] * in [i];

    for (i = 1; i <= n_in; i++) {
        f = -log (1.0 / in [i] - 1.0);
        f += eta * d0 [i];
        in [i] = 1.0 / (1.0 + exp (-f));
    }

    for (i = 0; i < in_vec_len; i++)
        if (
            (in_vecs [vectornumber][i] != nill)
         && (in_vecs [vectornumber][i] != any)
        )
            memcpy (
                vec_val [in_vecs [vectornumber][i]],
                in + 1 + i * vec_len,
                vec_len * sizeof (vec_val [0][0])
            );

    return 1;
}

float update_all_error ()       /*  return grootste fout: EXACT  */
{
    static float
        max_error;
    int
        i, j;

    if (all_updated)
        return max_error;

    max_error = 0;
    for (i = 0; i < n_pat; i++) {
        if (verbose == 2) {
            printf ("%3i\r", i);
            fflush (stdout);
        }
        feed_forward (i);
        for (j = 1; j <= n_out; j++)
            if ((p_err [i][j - 1] = fabs (out [j] - target [j])) > max_error)
                max_error = p_err [i][j - 1];
    }

    all_updated = 1;
    return max_error;
}

float display_error ()          /*  return grootste fout: BENADERING  */
{
    float
        high = 0,
        som = 0,
        w1gem,
        w1som,
        w1var,
        w2gem,
        w2som,
        w2var,
        vecgem,
        vecsom,
        vecvar;
    int
        i,
        j,
        n,
        n_min_1,
        er3 = 0;

    /*  bij gemiddelde van gewichten wordt bias niet gebruikt */
    n_min_1 = (n = n_in * n_hid) - 1;
    w1som = 0;
    for (i = 1; i <= n_in; i++)
        for (j = 1; j <= n_hid; j++)
            w1som += w1 [i][j];
    w1gem = w1som / n;
    w1var = 0;
    for (i = 1; i <= n_in; i++)
        for (j = 1; j <= n_hid; j++)
            w1var += pow (w1 [i][j] - w1gem, 2);
    w1var = w1var / n_min_1;

    n_min_1 = (n = n_hid * n_out) - 1;
    w2som = 0;
    for (i = 1; i <= n_hid; i++)
        for (j = 1; j <= n_out; j++)
            w2som += w2 [i][j];
    w2gem = w2som / n;
    w2var = 0;
    for (i = 1; i <= n_hid; i++)
        for (j = 1; j <= n_out; j++)
            w2var += pow (w2 [i][j] - w2gem, 2);
    w2var = w2var / n_min_1;

    for (i = 0; i < n_pat; i++)
        for (j = 0; j < n_out; j++) {
            if (high < p_err [i][j])
                high = p_err [i][j];
            som += p_err [i][j];
            if (p_err [i][j] > 3 * quit_val)
                er3++;
        }

    printf (
        "%6i%9.2f%9.3f%6i%8.3f%8.3f%8.3f%8.3f",
        epoch,
        som,
        high,
        er3,
        sqrt (w1var),
        sqrt (w2var),
        w1gem,
        w2gem
    );

    n = vec_len * (n_names - 2);    /*  niet !NILL en !ANY */
    n_min_1 = n - 1;

    vecsom = 0;
    for (i = 0; i < n_names; i++)
        if ((i != nill) && (i != any))
            for (j = 0; j < vec_len; j++)
                vecsom += vec_val [i][j];
    vecgem = vecsom / n;
    vecvar = 0;
    for (i = 0; i < n_names; i++)
        if ((i != nill) && (i != any))
            for (j = 0; j < vec_len; j++)
                vecvar += pow (vec_val [i][j] - vecgem, 2);
    vecvar = vecvar / n_min_1;

    printf ("%8.3f%8.3f\n", sqrt (vecvar), vecgem);
    fflush (stdout);

    return high;
}

void process_args (int argc, char *argv [])
{
    int
        i;
    char
        *illegal = "Ongeldige waarde voor %s";

    get_programname (argv [0]);

    if (argc == 1) {
        fprintf (
            stderr,
            "\nMiikkulainen's FGREP\n"
            "\n(c) P. Kleiweg 1996\n"
            "\nSyntax:\t%s -p string [-h int] [-v int] [-e float]\n"
            "\t[-w string] [-l string ]\n"
            "\t[-r float] [-n float] [-m int] [-s int]\n"
            "\t[-q float] [-x] [-z | -V]\n"
            "\n-p : naam bestand met patronen"
            "\n-h : aantal verborgen cellen"
            "\n-v : vectorlengte"
            "\n-e : eta (leerconstante)"
            "\n-w : naam bestand met gewichten, zoals bewaard door %s"
            "\n-l : naam bestand met waarden van vectoren, zoals bewaard door %s"
            "\n-r : maximale beginwaarde van de gewichten"
            "\n-n : maximale beginwaarde van de vectoren: 1 / (1 + e ^ -float)"
            "\n-m : maximaal aantal epochs"
            "\n-s : aantal epochs tussen uitvoer naar het scherm"
            "\n-q : quit wanneer grootste fout onder deze waarde komt"
            "\n-x : resultaten na training van netwerk niet bewaren"
            "\n-z : stil"
            "\n-V : verbose\n",
            programname,
            programname,
            programname
        );
        exit (1);
    }

    for (i = 1; i < argc; i++) {
        if (argv [i][0] != '-')
            errit ("Illegal argument");
        switch (argv [i][1]) {
            case 'p':
                patternfile = s_strdup (get_arg (argc, argv, &i));
                break;
            case 'h':
                n_hid = atoi (get_arg (argc, argv, &i));
                if (n_hid < 2)
                    errit (illegal, "-h");
                break;
            case 'v':
                vec_len = atoi (get_arg (argc, argv, &i));
                if (vec_len < 1)
                    errit (illegal, "-v");
                break;
            case 'e':
                eta = atof (get_arg (argc, argv, &i));
                if (eta <= 0.0)
                    errit (illegal, "-e");
                break;
            case 'w':
                oldweightsfile = s_strdup (get_arg (argc, argv, &i));
                break;
            case 'l':
                namesfile = s_strdup (get_arg (argc, argv, &i));
                break;
            case 'r':
                randomweight = atof (get_arg (argc, argv, &i));
                if (randomweight <= 0.0)
                    errit (illegal, "-r");
                break;
            case 'n':
                randomvector = atof (get_arg (argc, argv, &i));
                if (randomvector <= 0.0)
                    errit (illegal, "-n");
                break;
            case 'm':
                max_epoch = atoi (get_arg (argc, argv, &i));
                if (max_epoch < 0)
                    errit (illegal, "-m");
                break;
            case 's':
                show_epoch = atoi (get_arg (argc, argv, &i));
                if (show_epoch < 1)
                    errit (illegal, "-s");
                break;
            case 'q':
                quit_val = atof (get_arg (argc, argv, &i));
                if ((quit_val < 0.0) || (quit_val >= 1))
                    errit (illegal, "-q");
                break;
            case 'x':
                dont_save = 1;
                break;
            case 'z':
                verbose = 0;
                break;
            case 'V':
                verbose = 2;
                break;
            default:
                errit ("Illegal option '%s'", argv [i]);
        }
    }

    if (patternfile)
        read_vector_file ();
        /*  n_in en n_out zijn nu bekend */
    else
        errit ("Naam van file met patronen ontbreekt");

    if (namesfile)
        read_namesfile ();

    for (i = 0; i < vec_len; i++) {
        vec_val [nill][i] = 0.05;
        vec_val [any][i] = 0.5;
    }

    if (! n_hid)
        n_hid = n_in / 2;
    if (n_hid < 2)
        n_hid = 2;

    /*  het netwerk */
    in = (float *) s_malloc ((n_in + 1) * sizeof (float));
    hid = (float *) s_malloc ((n_hid + 1) * sizeof (float));
    out = (float *) s_malloc ((n_out + 1) * sizeof (float));
    target = (float *) s_malloc ((n_out + 1) * sizeof (float));
    in [0] = hid [0] = 1.0;

    d0 = (float *) s_malloc ((n_in + 1) * sizeof (float));

    w1 = (float **) s_malloc ((n_in + 1) * sizeof (float *));
    for (i = 0; i <= n_in; i++)
        w1 [i] = (float *) s_malloc ((n_hid + 1) * sizeof (float));

    d1 = (float *) s_malloc ((n_hid + 1) * sizeof (float));

    w2 = (float **) s_malloc ((n_hid + 1) * sizeof (float *));
    for (i = 0; i <= n_hid; i++)
        w2 [i] = (float *) s_malloc ((n_out + 1) * sizeof (float));
    d2 = (float *) s_malloc ((n_out + 1) * sizeof (float));

    if (oldweightsfile)
        read_oldweights ();
    else
        random_weights ();

    update_all_error ();

    if (verbose) {
        printf (
            " Eta:           %f\n"
            " Netwerk:       %i/%i/%i\n"
            " Patronen:      %s (%i)\n",
            eta,
            n_in,
            n_hid,
            n_out,
            patternfile,
            n_pat
        );

        printf (" Gewichten:     ");
        if (oldweightsfile)
            printf ("%s\n", oldweightsfile);
        else
            printf (
                "Random [%.2f, %.2f]\n",
                -randomweight,
                randomweight
            );

        printf (" Vectorlengte:  %i\n Vectoren:      ", vec_len);
        if (namesfile)
            printf ("%s\n", namesfile);
        else
            printf (
                "Random [%.2f, %.2f]\n",
                1.0 / (1.0 + exp (randomvector)),
                1.0 / (1.0 + exp (-randomvector))
            );

        printf (" Quit:          Max fout < %f", quit_val);
        printf (" of aantal rondes = %i\n", max_epoch);

        if (dont_save)
            printf (" Resultaten worden niet in een bestand bewaard\n");

#ifdef __MSDOS__
        printf (
            " Vrij geheugen: %lu bytes\n",
            (long unsigned) coreleft ()
        );
#endif
    }
}

char *get_arg (int argc, char *argv [], int *index)
{
    if (argv [*index][2])
        return argv [*index] + 2;

    if (*index == argc - 1)
        errit ("Argument voor optie '%s' ontbreekt", argv [*index]);

    return argv [++*index];
}

void save_results (int argc, char *argv [])
{
    int
        i,
        j,
        n;
    FILE
        *fp;
    float
        gem,
        som,
        var;
    char
        *file,
        *p;

    if (dont_save)
        return;

    file = (char *) s_malloc (strlen (patternfile) + 12);
    strcpy (file, patternfile);
    p = strrchr (file, PATH_SEP);
    if (p)
        p++;
    else
        p = file;

    for (i = 0; p [i]; i++)
        if (p [i] == '.')
            p [i] = '_';

    p [8] = '\0';
    while (strlen (p) < 8)
        strcat (p, "0");

    for (i = 0; ; i++) {
        sprintf (p + 5, "%03i", i);

        strcpy (p + 8, ".wt");
        if (access (file, 0) == 0)
            continue;
        strcpy (p + 8, ".log");
        if (access (file, 0) == 0)
            continue;
        strcpy (p + 8, ".vec");
        if (access (file, 0) == 0)
            continue;
        break;
    }

    strcpy (p + 8, ".wt");
    fp = fopen (file, "w");
    fprintf (fp, "# Net :\n %i %i %i\n", n_in, n_hid, n_out);
    for (i = 0; i <= n_in; i++) {
        fprintf (fp, "# In [%i] -> Hidden [1, %i] :\n", i, n_hid);
        for (j = 1; j <= n_hid; j++)
            fprintf (fp, "%16f\n", w1 [i][j]);
    }
    for (i = 0; i <= n_hid; i++) {
        fprintf (fp, "# Hidden [%i] -> Out [1, %i] :\n", i, n_out);
        for (j = 1; j <= n_out; j++)
            fprintf (fp, "%16f\n", w2 [i][j]);
    }
    fclose (fp);
    if (verbose)
        printf ("Gewichten bewaard in %s\n", file);

    strcpy (p + 8, ".vec");
    fp = fopen (file, "w");
    fprintf (fp, "%i\n", vec_len);
    for (i = 0; i < n_names; i++) {
        fprintf (fp, "%s\n", vec_names [i]);
        for (j = 0; j < vec_len; j++)
            fprintf (fp, "%16f\n", vec_val [i][j]);
    }
    fclose (fp);
    if (verbose)
        printf ("Vectoren bewaard in %s\n", file);

    strcpy (p + 8, ".log");
    fp = fopen (file, "w");
    fprintf (fp, "# %s", programname);
    for (i = 1; i < argc; i++)
        fprintf (fp, " %s", argv [i]);
    fprintf (fp, "\n# %i epochs\n", epoch);
    som = 0;
    n = 0;
    for (i = 0; i < n_pat; i++) {
        if (verbose == 2) {
            printf ("%3i\r", i);
            fflush (stdout);
        }
        for (j = 0; j < n_out; j++) {
            som += p_err [i][j];
            if (p_err [i][j] > quit_val)
                n++;
        }
    }

    gem = som / n_pat / n_out;
    var = 0;
    for (i = 0; i < n_pat; i++) {
        if (verbose == 2) {
            printf ("%3i\r", i);
            fflush (stdout);
        }
        for (j = 0; j < n_out; j++)
            var += pow (p_err [i][j] - gem, 2);
    }
    var = var / (n_pat * n_out - 1);
    fprintf (
        fp,
        "Fouten over %i patronen * %i uitvoercel%s:\n"
        "  Totaal:                   %g\n"
        "  Gemiddeld:                %g\n"
        "  Standaarddeviatie s(n-1): %g\n"
        "  Aantal groter dan %4.2f:   %i\n",
        n_pat,
        n_out,
        n_out > 1 ? "len" : "",
        som,
        gem,
        sqrt (var),
        quit_val,
        n
    );
/*  for (i = 0; i < n_pat; i++) {
        if (verbose == 2) {
            printf ("%3i\r", i);
            fflush (stdout);
        }
        for (j = 0; j < n_out; j++)
            if (p_err [i][j] > quit_val)
                break;
        if (j < n_out) {
            feed_forward (i);
            for (j = 0; j < in_vec_len; j++)
                fprintf (fp, "%s ", vec_names [in_vecs [i][j]]);
            fprintf (fp, " -> ");
            for (j = 0; j < out_vec_len; j++)
                fprintf (fp, " %s", vec_names [out_vecs [i][j]]);
            fprintf (fp, "\n  Tar:");
            for (j = 1; j <= n_in; j++)
                fprintf (fp, "%7.3f", in [j]);
            fprintf (fp, "\n  Out:");
            for (j = 1; j <= n_out; j++)
                fprintf (fp, "%8.4f", out [j]);
            fprintf (fp, "\n  Err:");
            for (j = 1; j <= n_out; j++)
                fprintf (fp, "%8.4f", fabs (target [j] - out [j]));
            fprintf (fp, "\n");
        }
    }
*/
    fclose (fp);
    if (verbose)
        printf ("Resultaten bewaard in %s\n", file);

    free (file);
}

void read_oldweights ()
{
    int
        nn_in, nn_hid, nn_out,
        i,
        j;

    FILE
        *fp = open_read (oldweightsfile);

    getline (fp, oldweightsfile);
    if (sscanf (buffer, " %u %u %u", &nn_in, &nn_hid, &nn_out) != 3)
        errit ("Error in %s", oldweightsfile);
    if ((nn_in != n_in) || (nn_hid != n_hid) || (nn_out != n_out))
        errit (
            "Invalid net (%u/%u/%u) in %s",
            nn_in,
            nn_hid,
            nn_out,
            oldweightsfile
        );

    for (i = 0; i <= n_in; i++)
        for (j = 1; j <= n_hid; j++) {
            getline (fp, oldweightsfile);
            if (sscanf (buffer, " %f", &(w1 [i][j])) != 1)
                errit ("Error in %s", oldweightsfile);
        }
    for (i = 0; i <= n_hid; i++)
        for (j = 1; j <= n_out; j++) {
            getline (fp, oldweightsfile);
            if (sscanf (buffer, " %f", &(w2 [i][j])) != 1)
                errit ("Error in %s", oldweightsfile);
        }
    fclose (fp);
}

void random_weights ()
{
    int
        i, j;
    float
        HRM;

    HRM = RAND_MAX / 2;
    for (i = 0; i <= n_in; i++)
        for (j = 1; j <= n_hid; j++)
            w1 [i][j] = (((float) rand ()) - HRM) / HRM * randomweight;
    for (i = 0; i <= n_hid; i++)
        for (j = 1; j <= n_out; j++)
            w2 [i][j] = (((float) rand ()) - HRM) / HRM * randomweight;
}

void rnd_vec ()
{
    int
        i, n, swap;
    for (i = n_pat - 1; i; i--) {
        n = rand () % (i + 1);
        swap = p_vec [n];
        p_vec [n] = p_vec [i];
        p_vec [i] = swap;
    }
}

void read_vector_file ()
{
    int
        i, j,
        p,
        n,
        offset,
        mi,
        mo,
        sl;
    char
        *buf;
    float
        f,
        HRM;
    FILE
        *fp;

    buf = (char *) s_malloc (256);

    max_names = 0;
    vec_names = 0;
    in_vec_len = 0;
    out_vec_len = 0;
    n_pat = 0;

    nill = locate ("!NILL");
    any = locate ("!ANY");
    fp = open_read (patternfile);
    while (fgets (buffer, BUFSIZE, fp)) {
        if ((buffer [0] == '\n') || (buffer [0] == '#'))
            continue;
        mi = mo = sl = offset = 0;
        for (;;) {
            if (sscanf (buffer + offset, " %s%n", buf, &n) == EOF)
                break;
            offset += n;
            if (! strcmp ("/", buf)) {
                if (sl)
                    errit ("Ongeldige vector: %s", buffer);
                sl = 1;
                continue;
            }
            locate (buf);
            if (sl)
                mo++;
            else
                mi++;
        }
        if (mi > in_vec_len)    /*  maximale lengte invoer  */
            in_vec_len = mi;
        if (mo > out_vec_len)   /*  maximale lengte uitvoer  */
            out_vec_len = mo;
        n_pat++;             /*  aantal gevonen invoer/uitvoer paren  */
    }
    fclose (fp);
    if ((! in_vec_len) || (! out_vec_len) || (! n_pat))
        errit ("Geen geldige vectoren gevonden");

    n_in = vec_len * in_vec_len;
    n_out = vec_len * out_vec_len;

    HRM = RAND_MAX / 2;

    vec_val = (float **) s_malloc (n_names * sizeof (float *));
    for (i = 0; i < n_names; i++) {
        vec_val [i] = (float *) s_malloc (vec_len * sizeof (float));
        for (j = 0; j < vec_len; j++) {
            f = (((float) rand ()) - HRM) / HRM * randomvector;
            vec_val [i][j] = 1.0 / (1.0 + exp (-f));
        }
    }

    in_vecs = (int **) s_malloc (n_pat * sizeof (int *));
    out_vecs = (int **) s_malloc (n_pat * sizeof (int *));
    p_err = (float **) s_malloc (n_pat * sizeof (float *));
    p_vec = (int *) s_malloc (n_pat  * sizeof (int));
    for (i = 0; i < n_pat; i++) {
        in_vecs [i] = (int *) s_malloc (in_vec_len * sizeof (int));
        for (j = 0; j < in_vec_len; j++)
            in_vecs [i][j] = nill;
        out_vecs [i] = (int *) s_malloc (out_vec_len * sizeof (int));;
        for (j = 0; j < out_vec_len; j++)
            out_vecs [i][j] = nill;
        p_err [i] = (float *) s_malloc (n_out * sizeof (float));;
        p_vec [i] = i;
    }

    p = 0;
    fp = open_read (patternfile);
    while (fgets (buffer, BUFSIZE, fp)) {
        if ((buffer [0] == '\n') || (buffer [0] == '#'))
            continue;
        mi = mo = sl = offset = 0;
        for (;;) {
            if (sscanf (buffer + offset, " %s%n", buf, &n) == EOF)
                break;
            offset += n;
            if (! strcmp ("/", buf)) {
                sl = 1;
                continue;
            }
            if (sl)
                out_vecs [p][mo++] = locate (buf);
            else
                in_vecs [p][mi++] = locate (buf);
        }
        p++;
    }
    fclose (fp);
    free (buf);
}

void read_namesfile ()
{
    int
        i,
        p,
        file_vec_len,
        stored_names = n_names;
    char
        *name;
    FILE
        *fp;

    name = (char *) s_malloc (256);

    fp = open_read (namesfile);
    getline (fp, namesfile);
    if (sscanf (buffer, " %u", &file_vec_len) != 1)
        errit ("Vectorlengte niet gevonden in %s", namesfile);
    if (file_vec_len != vec_len)
        errit ("Ongeldige vectorlengte (%u) in %s", file_vec_len, namesfile);
    while (fgets (buffer, BUFSIZE, fp)) {
        if (buffer [0] == '\n')
            continue;
        sscanf (buffer, " %s", name);
        p = locate (name);
        for (i = 0; i < vec_len; i++) {
            getline (fp, namesfile);
            if (p < stored_names)
                if (sscanf (buffer, " %f", &(vec_val [p][i])) != 1)
                    errit ("Waarde voor vector (%s) ontbreekt", name);
        }
    }
    fclose (fp);
    free (name);

    n_names = stored_names;
}

int locate (char const *name)
{
    int
        i;

    for (i = 0; i < n_names; i++)
        if (! stricmp (vec_names [i], name))
            return i;

    if (n_names >= max_names) {
        max_names += 64;
        vec_names =
            (char **) s_realloc (vec_names, max_names * sizeof (char *));
    }
    vec_names [n_names++] = s_strdup (name);
    return n_names - 1;
}


FILE *open_read (char const *filename)
{
    FILE *fp = fopen (filename, "r");
    if (! fp)
        errit ("Bestand niet gevonden: %s", filename);
    return fp;
}

char *getline (FILE *fp, char const *required)
/*  required is de naam van het bestand indien een regel MOET worden gevonden  */
{
    int i;
    while (fgets (buffer, BUFSIZE, fp)) {
        i = 0;
        while (buffer [i])
            switch (buffer [i++]) {
                case ' ':
                case '\t':
                case '\n':
                    break;
                case '#':
                    i = strlen (buffer);
                    break;
                default:
                    return buffer + i - 1;
            }
    }
    if (required) {
        fclose (fp);
        errit ("Onverwacht einde in file %s", required);
    }
    return 0;
}

void no_mem ()
{
    /*  geheugen vrijmaken  */
    free (no_mem_buffer);

    /*  foutmelding  */
    errit ("Out of memory!");
}

void *s_malloc (size_t size)
{
    void
        *p;

    p = malloc (size);
    if (! p) {
        free (no_mem_buffer);
        errit (Out_of_memory);
    }
    return p;
}

void *s_realloc (void *block, size_t size)
{
    void
        *p;

    p = realloc (block, size);
    if (! p) {
        free (no_mem_buffer);
        errit (Out_of_memory);
    }
    return p;
}

char *s_strdup (char const *s)
{
    char
        *s1;

    if (s) {
        s1 = (char *) s_malloc (strlen (s) + 1);
        strcpy (s1, s);
    } else {
        s1 = (char *) s_malloc (1);
        s1 [0] = '\0';
    }
    return s1;
}

void get_programname (char const *argv0)
{
#ifdef __MSDOS__
    char
        name [MAXFILE];
    fnsplit (argv0, NULL, NULL, name, NULL);
    programname = strdup (name);
#else   /* unix */
    char
        *p;
    p = strrchr (argv0, '/');
    if (p)
        programname = strdup (p + 1);
    else
        programname = strdup (argv0);
#endif    
}

void errit (char const *format, ...)
{
    va_list
        list;

    fprintf (stderr, "\nError %s: ", programname);

    va_start (list, format);
    vfprintf (stderr, format, list);

    fprintf (stderr, "\n\n");

    exit (1);
}
